==> scrubyt-0.2.3.gem: lib/scrubyt/core/scraping/filter.rb <==
module Scrubyt
  ##
  #=<tt>Filter out relevant pieces from the parent pattern</tt>
  #
  #A Scrubyt extractor is almost like a waterfall: water is pouring from the top until
  #it reaches the bottom. The biggest difference is that instead of water, a HTML 
  #document travels through the space.
  #
  #Of course Scrubyt would not make much sense if the same document would arrive at 
  #the bottom that was poured in at the top - since in this case we might use an 
  #indentity transformation (i.e. do nothing with the input) as well. 
  #
  #This is where filters came in: as they name says, they filter the stuff that is
  #pouring from above, to leave the interesting parts and discard the rest. 
  #The working of a filter will be explained most easily by the help of an example.
  #Let's consider that we would like to extract information from a webshop; Concretely
  #we are interested in the name of the items and the URL pointing to the image of the 
  #item.
  #
  #To accomplish this, first we select the items with the pattern item (a pattern is 
  #a logical grouping of fillters; see Pattern documentation) Then our new 
  #context is the result extracted by the 'item' pattern; For every 'item' pattern, further
  #extract the name and the image of the item; and finally, extract the href attribute
  #of the image. Let's see an illustration:
  #
  #   root             --> This pattern is called a 'root pattern', It is invisible to you
  #   |                    and basically it represents the document; it has no filters
  #   +-- item         --> Filter what's coming from above (the whole document) to get 
  #       |                relevant pieces of data (in this case webshop items)
  #       +-- name     --> Again, filter what's coming from above (a webshop item) and
  #       |                leave only item names after this operation
  #       +-- image    --> This time filter the image of the item
  #           |
  #           +-- href --> And finally, from the image elements, get the attribute 'href'
  class Filter
    #Type of the example this filter is extracted with
    
    #XPath example, like html/body/tr/td[1] etc.
    EXAMPLE_TYPE_XPATH = 0
    #String from the document, for example 'Canon EOS 300 D'.
    EXAMPLE_TYPE_STRING = 1
    #Image example, like 'http://www.rubyrailways.com/scrubyt.jpg'
    EXAMPLE_TYPE_IMAGE = 2
    #No example - the actual XPath is determined from the children XPaths (their LCA)
    EXAMPLE_TYPE_CHILDREN = 3
    #Regexp example, like /\d+@*\d+[a-z]/
    EXAMPLE_TYPE_REGEXP = 4
    
    attr_accessor :example_type, :parent_pattern, :temp_sink, 
                  :constraints, :xpath, :regexp, :example, :source, :sink

    def initialize(parent_pattern, example=nil, *args)
      @parent_pattern = parent_pattern
      #If the example type is not explicitly defined in the pattern definition,
      #try to determine it automatically from the example
      #@example_type = (args[0] == nil ? Filter.determine_example_type(example) :
      #                                  args[0][:example_type])
      #TODOOOOO correct this!
      @example_type = Filter.determine_example_type(example)
      @sink = []                  #output of a filter
      @source = []                #input of a filter                                        
      @example = example
      @xpath = nil #The xpath to evaluate this filter
      #temp sinks are used for the initial run when determining the XPaths for examples;
      @temp_sink = nil 
      @constraints = [] #list of constraints
    end
    
    #Evaluate this filter. This method shoulf not be called directly - as the pattern hierarchy 
    #is evaluated, every pattern evaluates its filters and then they are calling this method
    def evaluate(source)
      case @parent_pattern.type        
        when Scrubyt::Pattern::PATTERN_TYPE_TREE      
          result = source/@xpath
          #puts "Evaluating #{@parent_pattern.name} with #{@xpath}"
          result.class == Hpricot::Elements ? result.map : [result]
        when Scrubyt::Pattern::PATTERN_TYPE_ATTRIBUTE  
          attribute_value = [source.attributes[@example]] 
          return attribute_value if attribute_value[0]
          @@attribute_in_parent = nil
          Filter.traverse_up_until_attribute_found(source.parent, @example)
          @@attribute_in_parent
                  
        when Scrubyt::Pattern::PATTERN_TYPE_REGEXP
          source.inner_text.scan(@example).flatten
      end      
    end
             
    #For all the tree patterns, generate an XPath based on the given example
    #Also this method should not be called directly; It is automatically called for every tree
    #pattern directly after wrapper definition
    def generate_XPath_for_example
      case @example_type
        when EXAMPLE_TYPE_XPATH
          @xpath = @example
        when EXAMPLE_TYPE_STRING
          @temp_sink = XPathUtils.find_node_from_text( @parent_pattern.evaluation_context.root_pattern.filters[0].source[0], 
	                                                   @example, 
						                               false )
          @xpath = @parent_pattern.generalize ? XPathUtils.generate_XPath(@temp_sink, nil, false) :
                                                 XPathUtils.generate_XPath(@temp_sink, nil, true)
        when EXAMPLE_TYPE_CHILDREN          
          current_example_index = 0
          loop do
            all_child_temp_sinks = []
            @parent_pattern.children.each do |child_pattern|
              all_child_temp_sinks << child_pattern.filters[current_example_index].temp_sink
            end
            
            result = all_child_temp_sinks.pop
            if all_child_temp_sinks.empty?
              result = result.parent
            else
              all_child_temp_sinks.each do |child_sink|              
                result = XPathUtils.lowest_common_ancestor(result, child_sink)
              end
            end
            xpath = @parent_pattern.generalize ? XPathUtils.generate_XPath(result, nil, false) :
                                                   XPathUtils.generate_XPath(result, nil, true)
            if @parent_pattern.filters.size < current_example_index + 1
              @parent_pattern.filters << Scrubyt::Filter.new(@parent_pattern)
            end  
            @parent_pattern.filters[current_example_index].xpath = xpath
            @parent_pattern.filters[current_example_index].temp_sink = result
            @parent_pattern.children.each do |child_pattern|
                  child_pattern.filters[current_example_index].xpath = 
                    child_pattern.generalize ? XPathUtils.generate_generalized_relative_XPath(child_pattern.filters[current_example_index].temp_sink, result) :
                                               XPathUtils.generate_relative_XPath(child_pattern.filters[current_example_index].temp_sink, result)                                            
            end                                                
            if @parent_pattern.children[0].examples == nil
              break if @parent_pattern.children[0].filters.size == current_example_index+1
            else
              break if @parent_pattern.children[0].examples.size == current_example_index+1
            end             
            current_example_index += 1
          end
        when EXAMPLE_TYPE_IMAGE
          @temp_sink = XPathUtils.find_image(@parent_pattern.evaluation_context.root_pattern.filters[0].source[0], @example)
          @xpath = XPathUtils.generate_XPath(@temp_sink, nil, false)
      end
    end

    #Dispatcher method to add constraints; of course, as with any method_missing, this method
    #should not be called directly
    def method_missing(method_name, *args, &block)
      constraints << Constraint.send("add_#{method_name.to_s}".to_sym, self, *args)
    end    
    
private    
    def self.traverse_up_until_attribute_found(source, attribute)
      if (!source.parent.is_a? Hpricot::Doc)
        #p source.attributes
        #p attribute
        #p source.attributes[attribute]
        @@attribute_in_parent = source.attributes[attribute] if source.attributes[attribute]        
        traverse_up_until_attribute_found(source.parent, attribute) if !@attribute_in_parent
      end
    end
        
    def self.determine_example_type(example)
      if example.instance_of? Regexp
        EXAMPLE_TYPE_REGEXP
      else
        case example
          when nil
            EXAMPLE_TYPE_CHILDREN
          when /\.(jpg|png|gif|jpeg)$/
            EXAMPLE_TYPE_IMAGE
          when /^\/{1,2}[a-z]+\d?(\[\d+\])?(\/{1,2}[a-z]+\d?(\[\d+\])?)*$/
            (example.include? '/' || example.include?('[')) ? EXAMPLE_TYPE_XPATH : EXAMPLE_TYPE_STRING 
          else
            EXAMPLE_TYPE_STRING
        end
      end
    end #End of method determine_example_type
  end #End of class 
end #End of module

==> scrubyt-0.2.6.gem: lib/scrubyt/core/scraping/filter.rb <==
module Scrubyt
  ##
  #=<tt>Filter out relevant pieces from the parent pattern</tt>
  #
  #A Scrubyt extractor is almost like a waterfall: water is pouring from the top until
  #it reaches the bottom. The biggest difference is that instead of water, a HTML 
  #document travels through the space.
  #
  #Of course Scrubyt would not make much sense if the same document would arrive at 
  #the bottom that was poured in at the top - since in this case we might use an 
  #indentity transformation (i.e. do nothing with the input) as well. 
  #
  #This is where filters came in: as they name says, they filter the stuff that is
  #pouring from above, to leave the interesting parts and discard the rest. 
  #The working of a filter will be explained most easily by the help of an example.
  #Let's consider that we would like to extract information from a webshop; Concretely
  #we are interested in the name of the items and the URL pointing to the image of the 
  #item.
  #
  #To accomplish this, first we select the items with the pattern item (a pattern is 
  #a logical grouping of fillters; see Pattern documentation) Then our new 
  #context is the result extracted by the 'item' pattern; For every 'item' pattern, further
  #extract the name and the image of the item; and finally, extract the href attribute
  #of the image. Let's see an illustration:
  #
  #   root             --> This pattern is called a 'root pattern', It is invisible to you
  #   |                    and basically it represents the document; it has no filters
  #   +-- item         --> Filter what's coming from above (the whole document) to get 
  #       |                relevant pieces of data (in this case webshop items)
  #       +-- name     --> Again, filter what's coming from above (a webshop item) and
  #       |                leave only item names after this operation
  #       +-- image    --> This time filter the image of the item
  #           |
  #           +-- href --> And finally, from the image elements, get the attribute 'href'
  class Filter
    #Type of the example this filter is extracted with
    
    #XPath example, like html/body/tr/td[1] etc.
    EXAMPLE_TYPE_XPATH = 0
    #String from the document, for example 'Canon EOS 300 D'.
    EXAMPLE_TYPE_STRING = 1
    #Image example, like 'http://www.rubyrailways.com/scrubyt.jpg'
    EXAMPLE_TYPE_IMAGE = 2
    #No example - the actual XPath is determined from the children XPaths (their LCA)
    EXAMPLE_TYPE_CHILDREN = 3
    #Regexp example, like /\d+@*\d+[a-z]/
    EXAMPLE_TYPE_REGEXP = 4
    #Compound example, like :contains => 'goodies'
    EXAMPLE_TYPE_COMPOUND = 5
    
    attr_accessor :example_type, :parent_pattern, :temp_sink, 
                  :constraints, :xpath, :regexp, :example, :source, :sink

    def initialize(parent_pattern, example=nil, *args)
      @parent_pattern = parent_pattern
      #If the example type is not explicitly defined in the pattern definition,
      #try to determine it automatically from the example
      #@example_type = (args[0] == nil ? Filter.determine_example_type(example) :
      #                                  args[0][:example_type])
      #TODOOOOO correct this!
      @example_type = Filter.determine_example_type(example)
      @sink = []                  #output of a filter
      @source = []                #input of a filter                                        
      @example = example
      @xpath = nil #The xpath to evaluate this filter
      #temp sinks are used for the initial run when determining the XPaths for examples;
      #@temp_sink = nil 
      @constraints = [] #list of constraints
    end
    
    #Evaluate this filter. This method shoulf not be called directly - as the pattern hierarchy 
    #is evaluated, every pattern evaluates its filters and then they are calling this method
    def evaluate(source)
      case @parent_pattern.type        
        when Scrubyt::Pattern::PATTERN_TYPE_TREE      
          result = source/@xpath
          #puts "Evaluating #{@parent_pattern.name} with #{@xpath}"
          result.class == Hpricot::Elements ? result.map : [result]
        when Scrubyt::Pattern::PATTERN_TYPE_ATTRIBUTE  
          puts "Evaluating: #{@parent_pattern.name}"
          attribute_value = [source.attributes[@example]] 
          return attribute_value if attribute_value[0]
          @@attribute_in_parent = nil
          Filter.traverse_up_until_attribute_found(source.parent, @example)
          @@attribute_in_parent                  
        when Scrubyt::Pattern::PATTERN_TYPE_REGEXP
          source.inner_text.scan(@example).flatten
        when Scrubyt::Pattern::PATTERN_TYPE_DETAIL
          #p @parent_pattern.name
          result = @parent_pattern.evaluation_context.extractor.evaluate_subextractor(
            XPathUtils.find_nearest_node_with_attribute(source, 'href').attributes['href'],
            @parent_pattern)
      end      
    end
             
    #For all the tree patterns, generate an XPath based on the given example
    #Also this method should not be called directly; It is automatically called for every tree
    #pattern directly after wrapper definition
    def generate_XPath_for_example(next_page_example=false)
      #puts "generating example for: #{@parent_pattern.name}"
      #puts @example_type
      case @example_type
        when EXAMPLE_TYPE_XPATH
          @xpath = @example
        when EXAMPLE_TYPE_STRING
          @temp_sink = SimpleExampleLookup.find_node_from_text( @parent_pattern.evaluation_context.root_pattern.filters[0].source[0], 
	                                                            @example, 
						                                        next_page_example )
          @xpath = @parent_pattern.generalize ? XPathUtils.generate_XPath(@temp_sink, nil, false) :
                                                 XPathUtils.generate_XPath(@temp_sink, nil, true)
        when EXAMPLE_TYPE_CHILDREN
          current_example_index = 0
          loop do
            all_child_temp_sinks = []
            @parent_pattern.children.each do |child_pattern|
              all_child_temp_sinks << child_pattern.filters[current_example_index].temp_sink
            end
            result = all_child_temp_sinks.pop
            if all_child_temp_sinks.empty?
              result = result.parent
            else
              all_child_temp_sinks.each do |child_sink|              
                result = XPathUtils.lowest_common_ancestor(result, child_sink)
              end
            end
            xpath = @parent_pattern.generalize ? XPathUtils.generate_XPath(result, nil, false) :
                                                   XPathUtils.generate_XPath(result, nil, true)
            if @parent_pattern.filters.size < current_example_index + 1
              @parent_pattern.filters << Scrubyt::Filter.new(@parent_pattern)
            end  
            @parent_pattern.filters[current_example_index].xpath = xpath
            @parent_pattern.filters[current_example_index].temp_sink = result
            @parent_pattern.children.each do |child_pattern|            
                  next if child_pattern.type == Scrubyt::Pattern::PATTERN_TYPE_DETAIL
                  child_pattern.filters[current_example_index].xpath = 
                    child_pattern.generalize ? XPathUtils.generate_generalized_relative_XPath(child_pattern.filters[current_example_index].temp_sink, result) :
                                               XPathUtils.generate_relative_XPath(child_pattern.filters[current_example_index].temp_sink, result)                                            
            end                                                
            if @parent_pattern.children[0].examples == nil
              break if @parent_pattern.children[0].filters.size == current_example_index+1
            else
              break if @parent_pattern.children[0].examples.size == current_example_index+1
            end             
            current_example_index += 1
          end
        when EXAMPLE_TYPE_IMAGE
          @temp_sink = XPathUtils.find_image(@parent_pattern.evaluation_context.root_pattern.filters[0].source[0], @example)
          @xpath = XPathUtils.generate_XPath(@temp_sink, nil, false)
        when EXAMPLE_TYPE_COMPOUND        
          @temp_sink = CompoundExampleLookup.find_node_from_compund_example( @parent_pattern.evaluation_context.root_pattern.filters[0].source[0],
	                                                                         @example, 
						                                                     next_page_example )
          @xpath = @parent_pattern.generalize ? XPathUtils.generate_XPath(@temp_sink, nil, false) :
                                                 XPathUtils.generate_XPath(@temp_sink, nil, true)
      end
    end
    
    def setup_relative_XPaths
      return if !@parent_pattern.parent.parent
      parent_filter = @parent_pattern.parent.filters[@parent_pattern.filters.index(self)]
      @xpath = XPathUtils.generate_relative_XPath_from_XPaths(parent_filter.xpath, @xpath) if (@xpath =~ /^\/html/)
    end

    #Dispatcher method to add constraints; of course, as with any method_missing, this method
    #should not be called directly
    def method_missing(method_name, *args, &block)
      constraints << Constraint.send("add_#{method_name.to_s}".to_sym, self, *args)
    end    
    
private    
    def self.traverse_up_until_attribute_found(source, attribute)
      if (!source.parent.is_a? Hpricot::Doc)
        #p source.attributes
        #p attribute
        #p source.attributes[attribute]
        @@attribute_in_parent = source.attributes[attribute] if source.attributes[attribute]        
        traverse_up_until_attribute_found(source.parent, attribute) if !@attribute_in_parent
      end
    end
        
    def self.determine_example_type(example)
      if example.instance_of? Regexp
        EXAMPLE_TYPE_REGEXP
      elsif example.instance_of? Hash
        EXAMPLE_TYPE_COMPOUND
      else
        case example
          when nil
            EXAMPLE_TYPE_CHILDREN
          when /\.(jpg|png|gif|jpeg)$/
            EXAMPLE_TYPE_IMAGE
          when 
/^\/{1,2}[a-z]+[0-9]?(\[[0-9]+\])?(\/{1,2}[a-z()]+[0-9]?(\[[0-9]+\])?)*$/ 
            (example.include? '/' || example.include?('[')) ? EXAMPLE_TYPE_XPATH : EXAMPLE_TYPE_STRING 
          else
            EXAMPLE_TYPE_STRING
        end
      end
    end #End of method determine_example_type
  end #End of class 
end #End of module
